from collections import namedtuple

from django.urls import reverse_lazy as reverse
from django.utils.translation import ugettext_lazy as _

from ...forms.reports.courses import (
    CoursePaymentsForm, CoursePaymentsStatusForm, CourseStatsForm,
)
from ...forms.reports.debtors import DebtorsForm
from ...forms.reports.events import (
    EventPaymentsForm, EventPaymentsStatusForm, EventStatsForm,
)
from ...forms.reports.orderables import (
    OrderablePaymentsForm, OrderablePaymentsStatusForm, OrderableStatsForm,
)
from ..generic import TemplateView


class ReportsListView(TemplateView):
    template_name = 'leprikon/reports/index.html'

    Report = namedtuple('Report', ('title', 'instructions', 'form', 'url'))

    def get_context_data(self, *args, **kwargs):
        return super(ReportsListView, self).get_context_data(reports=[
            self.Report(
                title=_('Course payments'),
                instructions='',
                form=CoursePaymentsForm(),
                url=reverse('leprikon:report_course_payments'),
            ),
            self.Report(
                title=_('Event payments'),
                instructions='',
                form=EventPaymentsForm(),
                url=reverse('leprikon:report_event_payments'),
            ),
            self.Report(
                title=_('Orderable event payments'),
                instructions='',
                form=OrderablePaymentsForm(),
                url=reverse('leprikon:report_orderable_payments'),
            ),
            self.Report(
                title=_('Course payments status'),
                instructions='',
                form=CoursePaymentsStatusForm(),
                url=reverse('leprikon:report_course_payments_status'),
            ),
            self.Report(
                title=_('Event payments status'),
                instructions='',
                form=EventPaymentsStatusForm(),
                url=reverse('leprikon:report_event_payments_status'),
            ),
            self.Report(
                title=_('Orderable event payments status'),
                instructions='',
                form=OrderablePaymentsStatusForm(),
                url=reverse('leprikon:report_orderable_payments_status'),
            ),
            self.Report(
                title=_('Course statistics'),
                instructions='',
                form=CourseStatsForm(school_year=self.request.school_year),
                url=reverse('leprikon:report_course_stats'),
            ),
            self.Report(
                title=_('Event statistics'),
                instructions='',
                form=EventStatsForm(school_year=self.request.school_year),
                url=reverse('leprikon:report_event_stats'),
            ),
            self.Report(
                title=_('Orderable event statistics'),
                instructions='',
                form=OrderableStatsForm(school_year=self.request.school_year),
                url=reverse('leprikon:report_orderable_stats'),
            ),
            self.Report(
                title=_('Debtors list'),
                instructions='',
                form=DebtorsForm(),
                url=reverse('leprikon:report_debtors'),
            ),
        ])



from django.contrib import admin

from ..models.question import Question
from .filters import ActiveListFilter


@admin.register(Question)
class QuestionAdmin(admin.ModelAdmin):
    list_display = ('name', 'question', 'field')
    list_filter = (ActiveListFilter,)



from django import forms
from django.utils.translation import ugettext_lazy as _

from ...models.courses import Course
from ..form import FormMixin


class CoursePaymentsForm(FormMixin, forms.Form):
    date_start = forms.DateField(label=_('Start date'))
    date_end = forms.DateField(label=_('End date'))


class CoursePaymentsStatusForm(FormMixin, forms.Form):
    date = forms.DateField(label=_('Status for date'))


class CourseStatsForm(FormMixin, forms.Form):
    date = forms.DateField(label=_('Statistics for date'))
    courses = forms.ModelMultipleChoiceField(
        queryset=Course.objects.all(),
        label=_('Courses'),
    )
    paid_only = forms.BooleanField(label=_('Count only paid registrations'), required=False)

    def __init__(self, *args, **kwargs):
        school_year = kwargs.pop('school_year')
        super().__init__(*args, **kwargs)
        self.fields['courses'].queryset = Course.objects.filter(school_year=school_year)



from django.utils.translation import ugettext_lazy as _

from ..forms.billing_info import BillingInfoForm
from ..models.roles import BillingInfo
from ..utils import reverse_with_back
from .generic import CreateView, DeleteView, ListView, UpdateView


class GetQerysetMixin:
    def get_queryset(self):
        return self.request.user.leprikon_billing_info.all()


class BillingInfoListView(GetQerysetMixin, ListView):
    add_label = _('add billing information')
    preview_template = 'leprikon/billing_info_preview.html'

    def get_title(self):
        return _('Billing information')

    def get_add_url(self):
        return reverse_with_back(self.request, 'leprikon:billing_info_create')


class GetFromKwargsMixin:
    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs['user'] = self.request.user
        return kwargs


class BillingInfoCreateView(GetFromKwargsMixin, GetQerysetMixin, CreateView):
    form_class = BillingInfoForm
    title = _('New billing information')

    def get_message(self):
        return _('New billing information {} have been created.').format(self.object)


class BillingInfoUpdateView(GetFromKwargsMixin, GetQerysetMixin, UpdateView):
    form_class = BillingInfoForm
    message = _('Billing information have been updated.')
    title = _('Change billing information')


class BillingInfoDeleteView(GetQerysetMixin, DeleteView):
    model = BillingInfo
    title = _('Delete billing information')
    message = _('Billing information have been deleted.')

    def get_question(self):
        return _('Do You really want to delete the billing information: {}?').format(
            self.object
        )



from django.contrib import admin

from ..models.agegroup import AgeGroup


@admin.register(AgeGroup)
class AgeGroupAdmin(admin.ModelAdmin):
    list_display = ('name', 'order', 'min_age', 'max_age', 'require_school')
    list_editable = ('order', 'min_age', 'max_age', 'require_school')



from ..models.courses import CourseRegistration
from ..models.events import EventRegistration
from ..models.leprikonsite import LeprikonSite
from ..models.orderables import OrderableRegistration
from .generic import TemplateView


class RegistrationsListView(TemplateView):
    registrations = True
    template_name = 'leprikon/registrations.html'

    def get_context_data(self, **kwargs):
        context = super(RegistrationsListView, self).get_context_data(**kwargs)
        context['site'] = LeprikonSite.objects.get_current()
        context['courseregistrations'] = CourseRegistration.objects.filter(
            subject__school_year=self.request.school_year,
            user=self.request.user,
        )
        context['eventregistrations'] = EventRegistration.objects.filter(
            subject__school_year=self.request.school_year,
            user=self.request.user,
        )
        context['orderableregistrations'] = OrderableRegistration.objects.filter(
            subject__school_year=self.request.school_year,
            user=self.request.user,
        )
        return context



# -*- coding: utf-8 -*-
# Generated by Django 1.11.29 on 2020-03-22 08:36
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import leprikon.models.fields
import leprikon.models.subjects


class Migration(migrations.Migration):

    dependencies = [
        ('cms', '0022_auto_20180620_1551'),
        ('leprikon', '0049_target_groups_fix'),
    ]

    operations = [
        migrations.CreateModel(
            name='FilteredOrderableListPlugin',
            fields=[
                ('cmsplugin_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='leprikon_filteredorderablelistplugin', serialize=False, to='cms.CMSPlugin')),
            ],
            bases=('cms.cmsplugin',),
        ),
        migrations.CreateModel(
            name='Orderable',
            fields=[
                ('subject_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='leprikon.Subject')),
                ('duration', models.DurationField(help_text='Format: HH:MM:SS', verbose_name='duration')),
                ('due_from_days', models.IntegerField(blank=True, help_text='If set, payment request will be sent this number of days before event date. If not set, payment request will be sent when registration is approved.', null=True, verbose_name='number of days to send the payment request before event date')),
                ('due_date_days', models.IntegerField(default=0, verbose_name='number of days to due date before event date')),
            ],
            options={
                'verbose_name': 'orderable event',
                'verbose_name_plural': 'orderable events',
                'ordering': ('code', 'name'),
            },
            bases=('leprikon.subject',),
        ),
        migrations.CreateModel(
            name='OrderableDiscount',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('accounted', models.DateTimeField(default=django.utils.timezone.now, verbose_name='accounted time')),
                ('amount', leprikon.models.fields.PriceField(decimal_places=0, default=0, max_digits=10, verbose_name='discount')),
                ('explanation', models.CharField(blank=True, default='', max_length=250, verbose_name='discount explanation')),
            ],
            options={
                'verbose_name': 'orderable event discount',
                'verbose_name_plural': 'orderable event discounts',
                'ordering': ('accounted',),
            },
            bases=(leprikon.models.subjects.TransactionMixin, models.Model),
        ),
        migrations.CreateModel(
            name='OrderableListPlugin',
            fields=[
                ('cmsplugin_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='leprikon_orderablelistplugin', serialize=False, to='cms.CMSPlugin')),
                ('template', models.CharField(choices=[('default', 'Default'), ('grouped', 'Grouped by event groups')], default='default', help_text='The template used to render plugin.', max_length=100, verbose_name='template')),
                ('age_groups', models.ManyToManyField(blank=True, help_text='Keep empty to skip searching by age groups.', related_name='_orderablelistplugin_age_groups_+', to='leprikon.AgeGroup', verbose_name='age groups')),
                ('departments', models.ManyToManyField(blank=True, help_text='Keep empty to skip searching by departments.', related_name='_orderablelistplugin_departments_+', to='leprikon.Department', verbose_name='departments')),
            ],
            bases=('cms.cmsplugin',),
        ),
        migrations.CreateModel(
            name='OrderablePlugin',
            fields=[
                ('cmsplugin_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='leprikon_orderableplugin', serialize=False, to='cms.CMSPlugin')),
                ('template', models.CharField(choices=[('default', 'Default')], default='default', help_text='The template used to render plugin.', max_length=100, verbose_name='template')),
                ('event', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='leprikon.Orderable', verbose_name='event')),
            ],
            bases=('cms.cmsplugin',),
        ),
        migrations.CreateModel(
            name='OrderableRegistration',
            fields=[
                ('subjectregistration_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='leprikon.SubjectRegistration')),
                ('start_date', models.DateField(verbose_name='start date')),
                ('start_time', models.TimeField(blank=True, null=True, verbose_name='start time')),
            ],
            options={
                'verbose_name': 'orderable event registration',
                'verbose_name_plural': 'orderable event registrations',
            },
            bases=('leprikon.subjectregistration',),
        ),
        migrations.AlterField(
            model_name='subjecttype',
            name='subject_type',
            field=models.CharField(choices=[('course', 'course'), ('event', 'event'), ('orderable', 'orderable event')], max_length=10, verbose_name='subjects'),
        ),
        migrations.AddField(
            model_name='orderablelistplugin',
            name='event_types',
            field=models.ManyToManyField(blank=True, help_text='Keep empty to skip searching by event types.', related_name='_orderablelistplugin_event_types_+', to='leprikon.SubjectType', verbose_name='event types'),
        ),
        migrations.AddField(
            model_name='orderablelistplugin',
            name='groups',
            field=models.ManyToManyField(blank=True, help_text='Keep empty to skip searching by groups.', related_name='_orderablelistplugin_groups_+', to='leprikon.SubjectGroup', verbose_name='event groups'),
        ),
        migrations.AddField(
            model_name='orderablelistplugin',
            name='leaders',
            field=models.ManyToManyField(blank=True, help_text='Keep empty to skip searching by leaders.', related_name='_orderablelistplugin_leaders_+', to='leprikon.Leader', verbose_name='leaders'),
        ),
        migrations.AddField(
            model_name='orderablelistplugin',
            name='school_year',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='leprikon.SchoolYear', verbose_name='school year'),
        ),
        migrations.AddField(
            model_name='orderablelistplugin',
            name='target_groups',
            field=models.ManyToManyField(blank=True, help_text='Keep empty to skip searching by target groups.', related_name='_orderablelistplugin_target_groups_+', to='leprikon.TargetGroup', verbose_name='target groups'),
        ),
        migrations.AddField(
            model_name='orderablediscount',
            name='registration',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='discounts', to='leprikon.OrderableRegistration', verbose_name='registration'),
        ),
        migrations.AddField(
            model_name='filteredorderablelistplugin',
            name='event_types',
            field=models.ManyToManyField(related_name='_filteredorderablelistplugin_event_types_+', to='leprikon.SubjectType', verbose_name='event types'),
        ),
        migrations.AddField(
            model_name='filteredorderablelistplugin',
            name='school_year',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='leprikon.SchoolYear', verbose_name='school year'),
        ),
    ]



import base64
import hashlib
import re
import secrets
import uuid
from datetime import datetime
from itertools import chain

from django.contrib.auth import get_user_model
from django.db.models.signals import post_delete, post_save
from django.dispatch import receiver
from django.template.loader import get_template
from django.urls import reverse
from django.utils import translation
from django.utils.functional import cached_property
from django.utils.text import slugify
from pymongo import DESCENDING, MongoClient, UpdateMany, UpdateOne
from rocketchat_API.rocketchat import RocketChat as _RocketChat

from .conf import settings
from .models.leprikonsite import LeprikonSite
from .models.messages import Message
from .models.subjects import CHAT_GROUP_BROADCAST, Subject, SubjectRegistration

User = get_user_model()

http_re = re.compile('^http')

Leprikon = {'_id': 'leprikon', 'username': settings.LEPRIKON_CHAT_USERNAME}


def get_rc_url():
    return LeprikonSite.objects.get_current().url + '/_chat'


def get_rc_ws_url():
    return http_re.sub('ws', get_rc_url()) + '/websocket'


class QuerySet:
    def __init__(self, model, collection, query):
        self.model = model
        self._collection = collection
        self._query = query

    def _instance(self, doc):
        instance = self.model()
        instance.__dict__.update(doc)
        return instance

    def _get(self):
        obj = self._instance(self._collection.find_one(self._query))
        if not obj:
            raise self.model.DoesNotExist()
        return obj

    def _list(self):
        return [
            self._instance(doc)
            for doc in self._collection.find(
                self._query,
            ).sort(
                'ts',
                DESCENDING,
            )
        ]

    def __getitem__(self, key):
        return self._list().__getitem__(key)

    def __len__(self):
        return self._collection.count_documents(self._query)

    def _clone(self):
        return QuerySet(self.model, self._collection, self._query.copy())

    def count(self):
        return len(self)

    def filter(self, **kwargs):
        qs = self._clone()
        qs._query.update(kwargs)
        return qs

    def get(self, **kwargs):
        return self.filter(**kwargs)._get()


def get_rc_config():
    ls = LeprikonSite.objects.get_current()
    return (
        ('Accounts_iframe_enabled', True),
        ('Accounts_iframe_url', reverse('leprikon:user_login')),
        ('Accounts_Iframe_api_method', 'GET'),
        ('Accounts_Iframe_api_url', reverse('leprikon:api:rocketchat')),
        ('From_Email', settings.SERVER_EMAIL_PLAIN),
        ('Organization_Name', ls.name),
        ('Organization_Email', settings.SERVER_EMAIL_PLAIN),
        ('Organization_Type', 'nonprofit'),
        ('Country', 'czechRepublic'),
        ('Industry', 'education'),
        ('Language', 9),  # TODO: configurable language
        ('Layout_Sidenav_Footer', ''),
        ('Livechat_enabled', True),
        ('Server_Type', 'privateTeam'),
        ('Register_Server', False),
        ('SMTP_Host', settings.EMAIL_HOST),
        ('SMTP_Password', settings.EMAIL_HOST_PASSWORD),
        ('SMTP_Port', settings.EMAIL_PORT),
        ('SMTP_Username', settings.EMAIL_HOST_USER),
        ('Show_Setup_Wizard', 'completed'),
        ('Site_Name', ls.name),
        ('Site_Url', get_rc_url()),
        ('Website', ls.url + '/'),
    )


def get_rc_default_home(ls=None):
    if ls is None:
        ls = LeprikonSite.objects.get_current()
    with translation.override(settings.LANGUAGE_CODE):
        return get_template('rocketchat/home.html').render({
            'leprikon_site': ls,
        })


rc_role_permissions = (
    ('user', (
        'create-c',
        'create-d',
        'create-p',
        'create-personal-access-tokens',
        'delete-own-message',
        'leave-c',
        'leave-p',
        'mention-all',
        'mention-here',
        'view-c-room',
        'view-d-room',
        'view-history',
        'view-p-room',
        'preview-c-room',
        'view-outside-room',
        'start-discussion',
        'start-discussion-other-user',
    )),
    ('guest', (
        'create-d',
        'create-personal-access-tokens',
        'delete-own-message',
        'view-d-room',
        'view-history',
        'view-joined-room',
        'view-p-room',
        'start-discussion',
    )),
)


def get_rc_id(obj):
    if isinstance(obj, Subject):
        return f'subject-{obj.id}'
    return f'{obj.__class__.__name__.lower()}-{obj.id}'


def get_rc_u(user):
    return {'_id': get_rc_id(user), 'username': user.username}


def get_rc_leprikon_data():
    return {
        '_id': 'leprikon',
        'name': settings.LEPRIKON_CHAT_NAME,
        'username': settings.LEPRIKON_CHAT_USERNAME,
        'emails': [{
            'address': settings.SERVER_EMAIL_PLAIN,
            'verified': True,
        }],
        'active': True,
        'type': 'user',
        'roles': ['admin'],
        '__insert': {
            'createdAt': datetime.now(),
        },
    }


def get_rc_user_data(user):
    return {
        '_id': get_rc_id(user),
        'createdAt': user.date_joined,
        'name': user.get_full_name(),
        'username': user.username,
        'emails': [{'address': user.email, 'verified': True}],
        'active': user.is_active,
        'type': 'user',
        'roles': ['admin', 'livechat-agent'] if user.is_superuser else (
            ['user', 'livechat-agent'] if user.is_staff else ['guest']
        ),
        'lastLogin': user.last_login,
        '_updatedAt': datetime.now(),
    }


def get_value(document, key):
    parts = key.split('.', 1)
    value = document[parts[0]]
    return get_value(value, parts[1]) if len(parts) == 2 else value


def get_rc_subject_room_data(subject):
    fname = f'{subject} ({subject.id})'
    now = datetime.now()
    broadcast = (subject.chat_group_type or subject.subject_type.chat_group_type) == CHAT_GROUP_BROADCAST
    return {
        '_id': get_rc_id(subject),
        'name': slugify(fname.replace('/', '-')),
        'fname': fname,
        't': 'p',
        'ro': broadcast,
        'broadcast': broadcast,
        'sysMes': False,
        'default': False,
        'u': Leprikon,
        '_updatedAt': now,
        'customFields': {'leprikon': 'subject'},
        '__insert': {
            'msgs': 0,
            'usersCount': 0,
            'encrypted': False,
            'ts': now,
        },
    }


def get_rc_subscription_data(room_data, user, alert, groupMentions, owner_ids=None):
    now = datetime.now()
    subscription_data = {
        '_id': f'{room_data["_id"]}-{get_rc_id(user)}',
        'rid': room_data['_id'],
        'name': room_data['name'],
        'fname': room_data['fname'],
        't': room_data['t'],
        'u': get_rc_u(user),
        'customFields': {'leprikon': True},
        '_updatedAt': now,
        '__insert': {
            'alert': alert,
            'open': True,
            'ts': now,
            'ls': now,
            'unread': 0,
            'userMentions': 0,
            'groupMentions': groupMentions,
        },
    }
    if owner_ids:
        subscription_data['roles'] = ['owner'] if user.id in owner_ids else []
    return subscription_data


def get_rc_subject_subscriptions_data(room_data, subject):
    leader_users = {
        lu.user.id: lu.user
        for lu in subject.leaders.select_related('user')
    }
    return chain(
        get_rc_subscription_data(room_data, user, False, 0, leader_users)
        for user in set(chain(
            leader_users.values(),
            (r.user for r in subject.registrations.select_related('user')),
        ))
    )


def rc_logout(auth_token, user_id):
    return _RocketChat(
        auth_token=auth_token,
        user_id=user_id,
        server_url=settings.ROCKETCHAT_API_URL,
    ).logout().json()


class RocketChat:

    # mongo related helpers
    _mongo_session = None
    _mongo_trx_level = 0

    def __enter__(self):
        if not self._mongo_trx_level:
            self._mongo_session = self._mongo_client.start_session().__enter__()
            self._mongo_trx = self._mongo_session.start_transaction().__enter__()
        self._mongo_trx_level += 1
        return self

    def __exit__(self, *exc):
        self._mongo_trx_level -= 1
        if not self._mongo_trx_level:
            self._mongo_trx.__exit__(*exc)
            self._mongo_session.__exit__(*exc)
            self._mongo_session = None

    @cached_property
    def _mongo_client(self):
        return MongoClient(settings.MONGO_URL)

    @cached_property
    def _mongo(self):
        return self._mongo_client.get_default_database()

    def _get_query(self, document):
        on_insert = document.pop('__insert', False)
        query = {'$set': document}
        if on_insert:
            query['$setOnInsert'] = on_insert
        return query

    def _sync_one(self, collection, document, keys=('_id',)):
        return getattr(self._mongo, collection).update_one(
            {key: get_value(document, key) for key in keys},
            self._get_query(document),
            session=self._mongo_session,
            upsert=True,
        ).raw_result

    def _sync_many(self, collection, documents, keys=('_id',)):
        operations = [
            UpdateOne(
                {key: get_value(document, key) for key in keys},
                self._get_query(document),
                upsert=True,
            )
            for document in documents
        ]
        if operations:
            return getattr(self._mongo, collection).bulk_write(
                operations,
                session=self._mongo_session,
            ).bulk_api_result

    # RC proxy
    @cached_property
    def rc(self):
        self.sync_leprikon_user()
        return _RocketChat(
            user_id='leprikon',
            auth_token=self.create_login_token('leprikon'),
            server_url=settings.ROCKETCHAT_API_URL,
        )

    # Leprikon RC API
    def configure(self):
        return self._mongo.rocketchat_settings.bulk_write([
            UpdateOne({'_id': key}, {'$set': {'value': value, 'blocked': True}})
            for key, value in get_rc_config()
        ] + [
            # One time settings
            UpdateOne(
                {'_id': 'Layout_Home_Body', 'value': {'$regex': 'Welcome to Rocket.Chat'}},
                {'$set': {'value': get_rc_default_home()}},
            ),
        ], session=self._mongo_session).bulk_api_result

    def create_login_token(self, user_id):
        login_token = secrets.token_hex()
        h = hashlib.sha256()
        h.update(login_token.encode())
        hashed_token = base64.encodebytes(h.digest()).decode().strip()
        self._mongo.users.update_one(
            {'_id': user_id},
            {'$push': {
                'services.resume.loginTokens': {
                    'when': datetime.now(),
                    'hashedToken': hashed_token,
                }
            }},
            session=self._mongo_session,
        )
        return login_token

    def sync_leprikon_user(self):
        return self._sync_one('users', get_rc_leprikon_data())

    def sync_user(self, user):
        return self._sync_one('users', get_rc_user_data(user))

    def sync_users(self):
        return self._sync_many('users', map(get_rc_user_data, User.objects.iterator()))

    def sync_roles(self):
        return self._mongo.rocketchat_permissions.bulk_write(list(chain.from_iterable(
            (
                # add missing permissions
                UpdateMany(
                    {
                        '_id': {'$in': permissions},
                        'roles': {'$ne': role},
                    },
                    {'$push': {'roles': role}}
                ),
                # remove other permissions
                UpdateMany(
                    {
                        '_id': {'$nin': permissions},
                        'roles': role,
                    },
                    {'$pull': {'roles': role}}
                ),
            ) for role, permissions in rc_role_permissions
        )), session=self._mongo_session).bulk_api_result

    def sync_subject(self, subject):
        room_data = get_rc_subject_room_data(subject)
        with self:
            return (
                self._sync_one('rocketchat_room', room_data),
                self._sync_many(
                    'rocketchat_subscription',
                    get_rc_subject_subscriptions_data(room_data, subject)
                ),
            )

    def sync_subjects(self):
        room_data_subjects = [
            (get_rc_subject_room_data(subject), subject)
            for subject in Subject.objects.select_related('subject_type').iterator()
        ]
        with self:
            return (
                self._sync_many(
                    'rocketchat_room',
                    (rds[0] for rds in room_data_subjects)
                ),
                self._sync_many(
                    'rocketchat_subscription',
                    chain.from_iterable(
                        get_rc_subject_subscriptions_data(*rds)
                        for rds in room_data_subjects
                    ),
                ),
            )

    def sync_subscription(self, registration):
        room_data = get_rc_subject_room_data(registration.subject)
        return self._sync_one(
            'rocketchat_subscription',
            get_rc_subscription_data(room_data, registration.user, False, 0),
        )

    def merge_users(self, source, target):
        source_id = get_rc_id(source)
        target_id = get_rc_id(target)
        target_u = get_rc_u(target)
        with self:
            # merge rocketchat_livechat_department_agents
            for department_agent in self._mongo.rocketchat_livechat_department_agents.find({
                'agentId': source_id
            }, session=self._mongo_session):
                self._mongo.rocketchat_livechat_department_agents.count_documents({
                    'agentId': target_id,
                    'departmentId': department_agent['departmentId'],
                }) or self._sync_one(
                    'rocketchat_livechat_department_agents',
                    {
                        '_id': department_agent['_id'],
                        'agentId': target_id,
                        'username': target.username,
                    },
                )
            # merge rocketchat_subscription
            for subscription in self._mongo.rocketchat_subscription.find({
                'u._id': source_id
            }, session=self._mongo_session):
                subscription['_id'] = f'''{subscription['rid']}-{target_u['_id']}'''
                subscription['u'] = target_u
                self._sync_one(
                    'rocketchat_subscription',
                    subscription,
                )
            # merge transfer messages and rooms
            for collection in (
                self._mongo.rocketchat_message,
                self._mongo.rocketchat_room,
            ):
                collection.update_many(
                    {'u._id': source_id},
                    {'$set': {'u._id': target_id}},
                    session=self._mongo_session,
                )

    def get_leprikon_private_rooms_queryset(self):
        return QuerySet(Message, self._mongo.rocketchat_room, {
            't': 'p',
            'customFields.leprikon': {'$in': ['subject', 'broadcast']},
        })

    def get_broadcast_room_choices(self):
        return (
            (r['_id'], r['fname'])
            for r in self._mongo.rocketchat_room.find(
                {
                    't': 'p',
                    'broadcast': True,
                    'customFields.leprikon': 'broadcast',
                },
                ('_id', 'fname'),
            )
        )

    def subscribe_users_to_broadcast_room(self, rid, users):
        with self:
            room_data = self._mongo.rocketchat_room.find_one({
                '_id': rid,
                't': 'p',
                'broadcast': True,
                'customFields.leprikon': 'broadcast',
            }, session=self._mongo_session)
            self._sync_many(
                'rocketchat_subscription',
                (get_rc_subscription_data(room_data, user, True, 1) for user in users),
                keys=('rid', 'u._id'),
            )
            self._sync_one(
                'rocketchat_room',
                {
                    '_id': rid,
                    'usersCount': self._mongo.rocketchat_subscription.count_documents({'rid': rid}),
                },
            )

    def create_broadcast_room(self, author, subject, text, recipients):
        with self:
            name = slugify(subject.replace('/', '-'))
            now = datetime.now()
            room_data = {
                '_id': uuid.uuid4().hex[:10],
                'name': name,
                'fname': subject,
                't': 'p',
                'ro': True,
                'sysMes': False,
                'default': False,
                'u': get_rc_u(author),
                '_updatedAt': now,
                'msgs': 1,
                'usersCount': 0,
                'customFields': {'leprikon': 'broadcast'},
                'broadcast': True,
                'encrypted': False,
                'ts': now,
            }
            self._mongo.rocketchat_room.insert_one(room_data, session=self._mongo_session)
            self._mongo.rocketchat_message.insert_one({
                '_id': room_data['_id'],
                'rid': room_data['_id'],
                'msg': text,
                'ts': room_data['ts'],
                'u': room_data['u'],
                'mentions': [{
                    '_id': 'all',
                    'username': 'all',
                }],
                'channels': [],
                '_updatedAt': now,
            }, session=self._mongo_session)
            self.subscribe_users_to_broadcast_room(room_data['_id'], recipients)


@receiver(post_save)
def rc_sync_object(instance, **kwargs):
    if isinstance(instance, User):
        RocketChat().sync_user(instance)
    elif isinstance(instance, Subject):
        RocketChat().sync_subject(instance)
    elif isinstance(instance, LeprikonSite):
        RocketChat().configure()
    elif isinstance(instance, SubjectRegistration):
        RocketChat().sync_subscription(instance)


@receiver(post_delete)
def rc_delete_object(instance, **kwargs):
    if isinstance(instance, User):
        RocketChat().rc.users_delete(get_rc_id(instance))
    elif isinstance(instance, Subject):
        RocketChat().rc.channels_delete(get_rc_id(instance))



from datetime import date, datetime

import pytz
from django.http import HttpResponse, JsonResponse
from django.shortcuts import get_object_or_404
from django.utils.timezone import localdate

from ..models.subjects import Subject
from ..rocketchat import RocketChat, get_rc_id
from ..views import leader_or_staff_required


@leader_or_staff_required
def participants(request, subject_id):
    qs = Subject.objects.all()
    if not request.user.is_staff:
        qs = qs.filter(leaders=request.leader)
    subject = get_object_or_404(qs, id=subject_id)

    try:
        d = localdate(datetime.utcfromtimestamp(int(request.GET['date'])).replace(tzinfo=pytz.utc))
    except (KeyError, ValueError):
        d = date.today()

    return JsonResponse({
        'participants': list(
            {'value': p.id, 'label': str(p)}
            for p in subject.get_valid_participants(d)
        ),
    })


def rocketchat(request):
    """
    Implements API for Rocket.Chat IFrame authentication
    """
    if not request.user.is_authenticated():
        return HttpResponse(status=401)

    rc = RocketChat()
    rc.sync_user(request.user)

    return JsonResponse({
        'loginToken': rc.create_login_token(get_rc_id(request.user))
    })



from collections import namedtuple
from datetime import date, datetime, timedelta

from cms.models import CMSPlugin
from django.db import models, transaction
from django.dispatch import receiver
from django.utils.encoding import force_text
from django.utils.formats import date_format
from django.utils.functional import cached_property
from django.utils.translation import ugettext_lazy as _

from ..conf import settings
from ..utils import comma_separated
from .agegroup import AgeGroup
from .department import Department
from .fields import DAY_OF_WEEK, DayOfWeekField
from .journals import JournalEntry
from .roles import Leader
from .schoolyear import SchoolYear, SchoolYearDivision, SchoolYearPeriod
from .startend import StartEndMixin
from .subjects import (
    Subject, SubjectDiscount, SubjectGroup, SubjectRegistration, SubjectType,
)
from .targetgroup import TargetGroup
from .utils import PaymentStatus, change_year


class Course(Subject):
    school_year_division = models.ForeignKey(SchoolYearDivision, on_delete=models.PROTECT,
                                             related_name='courses', verbose_name=_('school year division'))
    allow_period_selection = models.BooleanField(
        _('allow period selection'),
        default=False,
        help_text=_('allow user to choose school year periods on registration form'),
    )

    class Meta:
        app_label = 'leprikon'
        ordering = ('code', 'name')
        verbose_name = _('course')
        verbose_name_plural = _('courses')

    @cached_property
    def all_times(self):
        return list(self.times.all())

    @cached_property
    def all_periods(self):
        return list(self.school_year_division.periods.all())

    @cached_property
    def all_journal_entries(self):
        return list(self.journal_entries.all())

    def get_times_list(self):
        return comma_separated(self.all_times)
    get_times_list.short_description = _('times')

    def get_next_time(self, now=None):
        try:
            return min(t.get_next_time(now) for t in self.all_times)
        except ValueError:
            return None

    @property
    def registrations_history_registrations(self):
        return CourseRegistration.objects.filter(course_history__course=self).distinct()

    @property
    def inactive_registrations(self):
        return (CourseRegistration.objects.filter(course_history__course=self)
                .exclude(id__in=self.active_registrations.all()).distinct())

    def copy_to_school_year(old, school_year):
        new = Course.objects.get(id=old.id)
        new.id, new.pk = None, None
        new.school_year = school_year
        new.public = False
        new.evaluation = ''
        new.note = ''
        year_delta = school_year.year - old.school_year.year
        new.school_year_division = SchoolYearDivision.objects.filter(
            school_year=school_year,
            name=old.school_year_division.name,
        ).first() or old.school_year_division.copy_to_school_year(school_year)
        new.reg_from = new.reg_from and change_year(new.reg_from, year_delta)
        new.reg_to = new.reg_to and change_year(new.reg_to, year_delta)
        new.save()
        new.times.set(old.times.all())
        new.groups.set(old.groups.all())
        new.age_groups.set(old.age_groups.all())
        new.target_groups.set(old.target_groups.all())
        for leader in old.all_leaders:
            school_year.leaders.add(leader)
        new.leaders.set(old.all_leaders)
        new.questions.set(old.questions.all())
        new.attachments.set(old.attachments.all())
        return new


class CourseTime(StartEndMixin, models.Model):
    course = models.ForeignKey(Course, on_delete=models.CASCADE,
                               related_name='times', verbose_name=_('course'))
    day_of_week = DayOfWeekField(_('day of week'))
    start = models.TimeField(_('start time'), blank=True, null=True)
    end = models.TimeField(_('end time'), blank=True, null=True)

    class Meta:
        app_label = 'leprikon'
        ordering = ('day_of_week', 'start')
        verbose_name = _('time')
        verbose_name_plural = _('times')

    def __str__(self):
        if self.start is not None and self.end is not None:
            return _('{day}, {start:%H:%M} - {end:%H:%M}').format(
                day=self.day,
                start=self.start,
                end=self.end,
            )
        elif self.start is not None:
            return _('{day}, {start:%H:%M}').format(
                day=self.day,
                start=self.start,
            )
        else:
            return force_text(self.day)

    @cached_property
    def day(self):
        return DAY_OF_WEEK[self.day_of_week]

    Time = namedtuple('Time', ('date', 'start', 'end'))

    def get_next_time(self, now=None):
        now = now or datetime.now()
        daydelta = (self.day_of_week - now.isoweekday()) % 7
        if daydelta == 0 and (isinstance(now, date) or self.start is None or self.start <= now.time()):
            daydelta = 7
        if isinstance(now, datetime):
            next_date = now.date() + timedelta(daydelta)
        else:
            next_date = now + timedelta(daydelta)
        return self.Time(
            date=next_date,
            start=self.start,
            end=self.end,
        )


class CourseRegistration(SubjectRegistration):
    subject_type = SubjectType.COURSE

    class Meta:
        app_label = 'leprikon'
        verbose_name = _('course registration')
        verbose_name_plural = _('course registrations')

    @cached_property
    def all_registration_periods(self):
        return list(
            self.course_registration_periods.select_related('period')
        )

    def get_period_payment_statuses(self, d=None):
        paid = self.get_paid(d)
        for counter, registration_period in enumerate(self.all_registration_periods, start=1):
            period_payment_status = registration_period.get_payment_status(
                paid=paid,
                last_period=counter == len(self.all_registration_periods),
                d=d,
            )
            yield period_payment_status
            paid -= period_payment_status.status.paid
        if not self.all_registration_periods:
            # create fake registration period to ensure some payment status
            registration_period = CourseRegistrationPeriod(
                registration=self,
                period=SchoolYearPeriod(
                    name=_('no period'),
                )
            )
            yield CourseRegistrationPeriod.PeriodPaymentStatus(
                period=registration_period.period,
                registration_period=registration_period,
                status=PaymentStatus(
                    price=0,
                    discount=0,
                    explanation='',
                    paid=paid,
                    current_date=d,
                    due_from=None,
                    due_date=None,
                ),
            )

    def get_payment_status(self, d=None):
        return sum(
            pps.status
            for pps in self.get_period_payment_statuses(d)
        )

    @cached_property
    def period_payment_statuses(self):
        return list(self.get_period_payment_statuses())

    @transaction.atomic
    def request_payment(self, payment_requested_by):
        self.course_registration_periods.filter(
            period__due_from__lte=date.today(),
        ).update(payment_requested=True)
        super().request_payment(payment_requested_by)


class CourseRegistrationPeriod(models.Model):
    registration = models.ForeignKey(
        CourseRegistration,
        on_delete=models.CASCADE,
        related_name='course_registration_periods',
        verbose_name=_('registration'),
    )
    period = models.ForeignKey(
        SchoolYearPeriod,
        on_delete=models.PROTECT,
        related_name='course_registration_periods',
        verbose_name=_('period'),
    )
    payment_requested = models.BooleanField(_('payment requested'), default=False)

    class Meta:
        app_label = 'leprikon'
        verbose_name = _('registered period')
        verbose_name_plural = _('registered periods')
        ordering = ('period__start',)

    def __str__(self):
        return str(self.period)

    @cached_property
    def all_discounts(self):
        # use cached self.registration.all_discounts instead of
        # return list(self.discounts.all())
        return [
            discount for discount in self.registration.all_discounts
            if discount.registration_period_id == self.id
        ]

    PeriodPaymentStatus = namedtuple(
        'PeriodPaymentStatus',
        ('period', 'registration_period', 'status'),
    )

    def get_payment_status(self, paid, last_period, d=None):
        if d is None:
            d = date.today()
        discount = sum(
            discount.amount
            for discount in self.all_discounts
            if discount.accounted.date() <= d
        )
        explanation = ',\n'.join(
            discount.explanation.strip()
            for discount in self.all_discounts
            if discount.accounted.date() <= d and discount.explanation.strip()
        )
        return self.PeriodPaymentStatus(
            period=self.period,
            registration_period=self,
            status=PaymentStatus(
                price=self.registration.price,
                discount=discount,
                explanation=explanation,
                paid=min(self.registration.price - discount, paid) if not last_period else paid,
                current_date=d,
                due_from=self.registration.payment_requested and max(
                    self.period.due_from,
                    self.registration.payment_requested.date(),
                ),
                due_date=self.registration.payment_requested and max(
                    self.period.due_date,
                    self.registration.payment_requested.date() + timedelta(
                        days=self.registration.subject.min_due_date_days
                    )
                ),
            ),
        )


class CourseDiscount(SubjectDiscount):
    registration = models.ForeignKey(
        CourseRegistration,
        on_delete=models.CASCADE,
        related_name='discounts',
        verbose_name=_('registration'),
    )
    registration_period = models.ForeignKey(
        CourseRegistrationPeriod,
        on_delete=models.CASCADE,
        related_name='discounts',
        verbose_name=_('period'),
    )

    @cached_property
    def period(self):
        return self.registration_period.period

    class Meta:
        app_label = 'leprikon'
        verbose_name = _('course discount')
        verbose_name_plural = _('course discounts')
        ordering = ('accounted',)


class CourseRegistrationHistory(StartEndMixin, models.Model):
    registration = models.ForeignKey(CourseRegistration, editable=False, on_delete=models.PROTECT,
                                     related_name='course_history', verbose_name=_('course'))
    course = models.ForeignKey(Course, editable=False, on_delete=models.PROTECT,
                               related_name='registrations_history', verbose_name=_('course'))
    start = models.DateField()
    end = models.DateField(blank=True, null=True)

    class Meta:
        app_label = 'leprikon'
        ordering = ('start',)
        verbose_name = _('course registration history')
        verbose_name_plural = _('course registration history')

    def __str__(self):
        return '{course}, {start} - {end}'.format(
            course=self.course.name,
            start=date_format(self.start, 'SHORT_DATE_FORMAT'),
            end=date_format(self.end, 'SHORT_DATE_FORMAT') if self.end else _('now'),
        )

    @property
    def journal_entries(self):
        qs = JournalEntry.objects.filter(subject_id=self.course_id, date__gte=self.start)
        if self.end:
            return qs.filter(date__lte=self.end)
        else:
            return qs

    def save(self, *args, **kwargs):
        if self.id:
            original = self.__class__.objects.get(id=self.id)
            min_journal_date = original.course_journal_entries.aggregate(models.Min('date'))['date__min']
            max_journal_date = original.course_journal_entries.aggregate(models.Max('date'))['date__max']
            # if journal entry exists, start must be set and lower or equal to min journal date
            if min_journal_date and self.start > min_journal_date:
                self.start = min_journal_date
            # end can not be lower than max journal date
            if self.end and max_journal_date and self.end < max_journal_date:
                self.end = max_journal_date
        super(CourseRegistrationHistory, self).save(*args, **kwargs)


@receiver(models.signals.post_save, sender=CourseRegistration)
def update_course_registration_history(sender, instance, created, **kwargs):
    if instance.approved:
        d = date.today()
        # if created or changed
        if (instance.course_history.count() == 0 or
                instance.course_history.filter(end=None).exclude(course_id=instance.subject_id).update(end=d)):
            # reopen or create entry starting today
            (
                CourseRegistrationHistory.objects.filter(
                    registration_id=instance.id, course_id=instance.subject_id, start=d,
                ).update(end=None) or
                CourseRegistrationHistory.objects.create(
                    registration_id=instance.id, course_id=instance.subject_id, start=d,
                )
            )


class CoursePlugin(CMSPlugin):
    course = models.ForeignKey(Course, on_delete=models.CASCADE, related_name='+', verbose_name=_('course'))
    template = models.CharField(_('template'), max_length=100,
                                choices=settings.LEPRIKON_COURSE_TEMPLATES,
                                default=settings.LEPRIKON_COURSE_TEMPLATES[0][0],
                                help_text=_('The template used to render plugin.'))

    class Meta:
        app_label = 'leprikon'


class CourseListPlugin(CMSPlugin):
    school_year = models.ForeignKey(SchoolYear, blank=True, null=True, on_delete=models.CASCADE,
                                    related_name='+', verbose_name=_('school year'))
    departments = models.ManyToManyField(Department, blank=True, related_name='+', verbose_name=_('departments'),
                                         help_text=_('Keep empty to skip searching by departments.'))
    course_types = models.ManyToManyField(SubjectType, blank=True,
                                          limit_choices_to={'subject_type': SubjectType.COURSE},
                                          related_name='+', verbose_name=_('course types'),
                                          help_text=_('Keep empty to skip searching by course types.'))
    age_groups = models.ManyToManyField(AgeGroup, blank=True, related_name='+', verbose_name=_('age groups'),
                                        help_text=_('Keep empty to skip searching by age groups.'))
    target_groups = models.ManyToManyField(TargetGroup, blank=True, related_name='+', verbose_name=_('target groups'),
                                           help_text=_('Keep empty to skip searching by target groups.'))
    groups = models.ManyToManyField(SubjectGroup, blank=True, related_name='+', verbose_name=_('course groups'),
                                    help_text=_('Keep empty to skip searching by groups.'))
    leaders = models.ManyToManyField(Leader, blank=True, related_name='+', verbose_name=_('leaders'),
                                     help_text=_('Keep empty to skip searching by leaders.'))
    template = models.CharField(_('template'), max_length=100,
                                choices=settings.LEPRIKON_COURSELIST_TEMPLATES,
                                default=settings.LEPRIKON_COURSELIST_TEMPLATES[0][0],
                                help_text=_('The template used to render plugin.'))

    class Meta:
        app_label = 'leprikon'

    def copy_relations(self, oldinstance):
        self.departments.set(oldinstance.departments.all())
        self.course_types.set(oldinstance.course_types.all())
        self.groups.set(oldinstance.groups.all())
        self.age_groups.set(oldinstance.age_groups.all())
        self.target_groups.set(oldinstance.target_groups.all())
        self.leaders.set(oldinstance.leaders.all())

    @cached_property
    def all_departments(self):
        return list(self.departments.all())

    @cached_property
    def all_course_types(self):
        return list(self.course_types.all())

    @cached_property
    def all_groups(self):
        return list(self.groups.all())

    @cached_property
    def all_age_groups(self):
        return list(self.age_groups.all())

    @cached_property
    def all_target_groups(self):
        return list(self.target_groups.all())

    @cached_property
    def all_leaders(self):
        return list(self.leaders.all())

    Group = namedtuple('Group', ('group', 'objects'))

    def render(self, context):
        school_year = (self.school_year or getattr(context.get('request'), 'school_year') or
                       SchoolYear.objects.get_current())
        courses = Course.objects.filter(school_year=school_year, public=True).distinct()

        if self.all_departments:
            courses = courses.filter(department__in=self.all_departments)
        if self.all_course_types:
            courses = courses.filter(subject_type__in=self.all_course_types)
        if self.all_age_groups:
            courses = courses.filter(age_groups__in=self.all_age_groups)
        if self.all_target_groups:
            courses = courses.filter(target_groups__in=self.all_target_groups)
        if self.all_leaders:
            courses = courses.filter(leaders__in=self.all_leaders)
        if self.all_groups:
            courses = courses.filter(groups__in=self.all_groups)
            groups = self.all_groups
        elif self.all_course_types:
            groups = SubjectGroup.objects.filter(subject_types__in=self.all_course_types)
        else:
            groups = SubjectGroup.objects.all()

        context.update({
            'school_year': school_year,
            'courses': courses,
            'groups': (
                self.Group(group=group, objects=courses.filter(groups=group))
                for group in groups
            ),
        })
        return context


class FilteredCourseListPlugin(CMSPlugin):
    school_year = models.ForeignKey(SchoolYear, blank=True, null=True, on_delete=models.CASCADE,
                                    related_name='+', verbose_name=_('school year'))
    course_types = models.ManyToManyField(SubjectType, limit_choices_to={'subject_type': SubjectType.COURSE},
                                          related_name='+', verbose_name=_('course types'))

    class Meta:
        app_label = 'leprikon'

    def copy_relations(self, oldinstance):
        self.course_types.set(oldinstance.course_types.all())

    @cached_property
    def all_course_types(self):
        return list(self.course_types.all())

    def render(self, context):
        school_year = (self.school_year or getattr(context.get('request'), 'school_year') or
                       SchoolYear.objects.get_current())

        from ..forms.subjects import SubjectFilterForm
        form = SubjectFilterForm(
            subject_type_type=SubjectType.COURSE,
            subject_types=self.all_course_types,
            school_year=school_year,
            is_staff=context['request'].user.is_staff,
            data=context['request'].GET,
        )
        context.update({
            'school_year': school_year,
            'form': form,
            'courses': form.get_queryset(),
        })
        return context



from datetime import date

from cms.models import CMSPlugin
from django.db import models
from django.utils.functional import cached_property
from django.utils.translation import ugettext_lazy as _

from ..utils import comma_separated
from .startend import StartEndMixin
from .utils import change_year


class SchoolYearManager(models.Manager):
    def get_current(self):
        # by default use last active year
        school_year = self.filter(active=True).order_by('-year').first()
        if school_year is None:
            # Create or activate current year
            if date.today().month < 7:
                year = date.today().year - 1
            else:
                year = date.today().year
            school_year = SchoolYear.objects.get_or_create(year=year)[0]
            school_year.active = True
            school_year.save()
        return school_year


class SchoolYear(models.Model):
    year = models.IntegerField(_('year'), unique=True, help_text=_('Only the first of the two years.'))
    active = models.BooleanField(_('active'), default=False)

    objects = SchoolYearManager()

    class Meta:
        app_label = 'leprikon'
        ordering = ('-year',)
        verbose_name = _('school year')
        verbose_name_plural = _('school years')

    def __str__(self):
        return self.name

    def delete(self):
        if self.courses.count() or self.events.count():
            raise Exception(_('Can not delete sclool year with courses or events.'))
        super(SchoolYear, self).delete()

    @cached_property
    def name(self):
        return '{}/{}'.format(self.year, self.year + 1)


class SchoolYearDivision(models.Model):
    school_year = models.ForeignKey(SchoolYear, on_delete=models.CASCADE,
                                    related_name='divisions', verbose_name=_('school year'))
    name = models.CharField(_('division name'), max_length=150)
    period_name = models.CharField(_('period name'), max_length=150)

    class Meta:
        app_label = 'leprikon'
        ordering = ('name',)
        unique_together = (('school_year', 'name'),)
        verbose_name = _('school year division')
        verbose_name_plural = _('school year divisions')

    def __str__(self):
        return self.name

    @cached_property
    def all_periods(self):
        return list(self.periods.all())

    def copy_to_school_year(old, school_year):
        new = SchoolYearDivision.objects.get(id=old.id)
        new.id, new.pk = None, None
        new.school_year = school_year
        new.save()
        year_delta = school_year.year - old.school_year.year
        periods = []
        for period in old.all_periods:
            periods.append(SchoolYearPeriod(
                school_year_division=new,
                name=period.name,
                start=change_year(period.start, year_delta),
                end=change_year(period.end, year_delta),
                due_from=change_year(period.due_from, year_delta),
                due_date=change_year(period.due_date, year_delta),
            ))
        SchoolYearPeriod.objects.bulk_create(periods)
        return new

    def get_current_period(self):
        return (
            self.periods.filter(end__gte=date.today()).first() or
            self.periods.last()
        )


class SchoolYearPeriod(StartEndMixin, models.Model):
    school_year_division = models.ForeignKey(SchoolYearDivision, on_delete=models.CASCADE,
                                             related_name='periods', verbose_name=_('school year division'))
    name = models.CharField(_('name'), max_length=150)
    start = models.DateField(_('start date'))
    end = models.DateField(_('end date'))
    due_from = models.DateField(_('due from'))
    due_date = models.DateField(_('due date'))

    class Meta:
        app_label = 'leprikon'
        ordering = ('start',)
        verbose_name = _('school year period')
        verbose_name_plural = _('school year periods')

    def __str__(self):
        return _('{name}, {start:%m/%d %y} - {end:%m/%d %y}').format(
            name=self.name,
            start=self.start,
            end=self.end,
        )


class SchoolYearBlockPlugin(CMSPlugin):
    school_years = models.ManyToManyField(
        SchoolYear,
        related_name='+',
        verbose_name=_('school years'),
        help_text=_('Which school years should the block be visible for?'),
    )

    class Meta:
        app_label = 'leprikon'

    def __str__(self):
        return comma_separated(self.all_school_years)

    @cached_property
    def all_school_years(self):
        return list(self.school_years.all())

    def copy_relations(self, oldinstance):
        self.school_years.set(oldinstance.school_years.all())



from . import (  # noqa
    account, agegroup, agreements, citizenship, courses, department, events,
    fields, journals, leprikonsite, messages, orderables, organizations, place,
    printsetup, question, registrationlink, roles, school, schoolyear,
    startend, subjects, targetgroup, timesheets, useragreement, utils,
)



from django import forms
from django.contrib import admin
from django.contrib.admin.widgets import AdminRadioSelect
from django.contrib.auth import get_user_model
from django.contrib.staticfiles.templatetags.staticfiles import static
from django.shortcuts import render
from django.urls import reverse
from django.utils.html import format_html
from django.utils.translation import ugettext_lazy as _

from ..models.orderables import (
    Orderable, OrderableDiscount, OrderableRegistration,
)
from ..models.schoolyear import SchoolYear
from ..models.subjects import SubjectType
from ..utils import currency
from .pdf import PdfExportAdminMixin
from .subjects import (
    SubjectBaseAdmin, SubjectPaymentBaseAdmin, SubjectRegistrationBaseAdmin,
)


@admin.register(Orderable)
class OrderableAdmin(SubjectBaseAdmin):
    subject_type_type = SubjectType.ORDERABLE
    registration_model = OrderableRegistration
    list_display = (
        'id', 'code', 'name', 'subject_type', 'get_groups_list', 'get_leaders_list',
        'duration',
        'place', 'public', 'registration_allowed_icon',
        'get_registrations_link',
        'get_journal_link', 'icon', 'note',
    )
    list_export = (
        'id', 'school_year', 'code', 'name', 'department', 'subject_type', 'registration_type',
        'get_groups_list', 'get_leaders_list', 'get_age_groups_list', 'get_target_groups_list',
        'place', 'public', 'price',
        'min_participants_count', 'max_participants_count', 'min_group_members_count', 'max_group_members_count',
        'min_registrations_count', 'max_registrations_count',
        'get_approved_registrations_count', 'get_unapproved_registrations_count', 'note',
    )
    actions = (
        'publish', 'unpublish',
        'copy_to_school_year',
    )

    def publish(self, request, queryset):
        Orderable.objects.filter(id__in=[reg['id'] for reg in queryset.values('id')]).update(public=True)
        self.message_user(request, _('Selected orderable events were published.'))
    publish.short_description = _('Publish selected orderable events')

    def unpublish(self, request, queryset):
        Orderable.objects.filter(id__in=[reg['id'] for reg in queryset.values('id')]).update(public=False)
        self.message_user(request, _('Selected orderable events were unpublished.'))
    unpublish.short_description = _('Unpublish selected orderable events')

    def copy_to_school_year(self, request, queryset):
        class SchoolYearForm(forms.Form):
            school_year = forms.ModelChoiceField(
                label=_('Target school year'),
                help_text=_('All selected orderable events will be copied to selected school year.'),
                queryset=SchoolYear.objects.all(),
            )
        if request.POST.get('post', 'no') == 'yes':
            form = SchoolYearForm(request.POST)
            if form.is_valid():
                school_year = form.cleaned_data['school_year']
                for orderable in queryset.all():
                    orderable.copy_to_school_year(school_year)
                self.message_user(
                    request,
                    _('Selected orderable events were copied to school year {}.').format(school_year),
                )
                return
        else:
            form = SchoolYearForm()

        adminform = admin.helpers.AdminForm(
            form,
            [(None, {'fields': list(form.base_fields)})],
            {},
            None,
            model_admin=self,
        )

        return render(request, 'leprikon/admin/action_form.html', dict(
            title=_('Select target school year'),
            opts=self.model._meta,
            adminform=adminform,
            media=self.media + adminform.media,
            action='copy_to_school_year',
            action_checkbox_name=admin.helpers.ACTION_CHECKBOX_NAME,
            select_across=request.POST['select_across'],
            selected=request.POST.getlist(admin.helpers.ACTION_CHECKBOX_NAME),
        ))
    copy_to_school_year.short_description = _('Copy selected orderable events to another school year')

    def get_message_recipients(self, request, queryset):
        return get_user_model().objects.filter(
            leprikon_registrations__subject__in=queryset,
        ).distinct()


@admin.register(OrderableRegistration)
class OrderableRegistrationAdmin(PdfExportAdminMixin, SubjectRegistrationBaseAdmin):
    subject_type_type = SubjectType.ORDERABLE
    actions = ('add_discounts',)
    list_display = (
        'variable_symbol', 'download_tag', 'subject_name', 'event_date', 'participants_list_html', 'price',
        'orderable_discounts', 'orderable_payments',
        'created_with_by', 'approved_with_by', 'payment_requested_with_by',
        'cancelation_requested_with_by', 'canceled_with_by',
        'note', 'random_number',
    )

    def add_discounts(self, request, queryset):
        ABSOLUTE = 'A'
        RELATIVE = 'R'

        class DiscountForm(forms.Form):
            discount_type = forms.ChoiceField(
                label=_('Discount type'),
                choices=(
                    (ABSOLUTE, _('absolute amount')),
                    (RELATIVE, _('relative amount in percents'))
                ),
                widget=AdminRadioSelect(),
            )
            amount = forms.DecimalField(label=_('Amount or number of percents'))
            explanation = OrderableDiscount._meta.get_field('explanation').formfield()

        if request.POST.get('post', 'no') == 'yes':
            form = DiscountForm(request.POST)
            if form.is_valid():
                OrderableDiscount.objects.bulk_create(
                    OrderableDiscount(
                        registration=registration,
                        amount=(
                            form.cleaned_data['amount']
                            if form.cleaned_data['discount_type'] == ABSOLUTE
                            else (
                                registration.price - sum(
                                    discount.amount
                                    for discount in registration.all_discounts
                                )
                            ) * form.cleaned_data['amount'] / 100
                        ),
                        explanation=form.cleaned_data['explanation'],
                    )
                    for registration in queryset.all()
                )
                self.message_user(request, _(
                    'The discounts have been created for selected registrations.'
                ))
                return
        else:
            form = DiscountForm()

        adminform = admin.helpers.AdminForm(
            form,
            [(None, {'fields': list(form.base_fields)})],
            {},
            None,
            model_admin=self,
        )

        return render(request, 'leprikon/admin/action_form.html', dict(
            title=_('Add discounts'),
            opts=self.model._meta,
            adminform=adminform,
            media=self.media + adminform.media,
            action='add_discounts',
            action_checkbox_name=admin.helpers.ACTION_CHECKBOX_NAME,
            select_across=request.POST['select_across'],
            selected=request.POST.getlist(admin.helpers.ACTION_CHECKBOX_NAME),
        ))
    add_discounts.short_description = _('Add discounts to selected registrations')

    def orderable_discounts(self, obj):
        return format_html(
            '<a href="{href_list}"><b>{amount}</b></a>'
            ' &nbsp; <a class="popup-link" href="{href_add}" style="background-position: 0 0" title="{title_add}">'
            '<img src="{icon_add}" alt="+"/></a>',
            href_list=reverse('admin:leprikon_orderablediscount_changelist') + '?registration={}'.format(obj.id),
            amount=currency(obj.payment_status.discount),
            href_add=reverse('admin:leprikon_orderablediscount_add') + '?registration={}'.format(obj.id),
            icon_add=static('admin/img/icon-addlink.svg'),
            title_add=_('add discount'),
        )
    orderable_discounts.allow_tags = True
    orderable_discounts.short_description = _('orderable event discounts')

    def orderable_payments(self, obj):
        return format_html(
            '<a class="popup-link" style="color: {color}" href="{href_list}" title="{title}"><b>{amount}</b></a>'
            ' &nbsp; <a class="popup-link" href="{href_add}" style="background-position: 0 0" title="{title_add}">'
            '<img src="{icon_add}" alt="+"/></a>',
            color=obj.payment_status.color,
            href_list=reverse('admin:leprikon_subjectpayment_changelist') + '?registration={}'.format(obj.id),
            title=obj.payment_status.title,
            amount=currency(obj.payment_status.paid),
            href_add=reverse('admin:leprikon_subjectpayment_add') + '?registration={}'.format(obj.id),
            icon_add=static('admin/img/icon-addlink.svg'),
            title_add=_('add payment'),
        )
    orderable_payments.allow_tags = True
    orderable_payments.short_description = _('orderable event payments')


@admin.register(OrderableDiscount)
class OrderableDiscountAdmin(PdfExportAdminMixin, SubjectPaymentBaseAdmin):
    list_display = ('accounted', 'registration', 'subject', 'amount_html', 'explanation')
    list_export = ('accounted', 'registration', 'subject', 'amount', 'explanation')



from django import forms
from django.contrib import admin
from django.db import IntegrityError
from django.shortcuts import render
from django.utils.translation import ugettext_lazy as _

from ..models.schoolyear import (
    SchoolYear, SchoolYearDivision, SchoolYearPeriod,
)
from .filters import SchoolYearListFilter


@admin.register(SchoolYear)
class SchoolYearAdmin(admin.ModelAdmin):
    list_display = ('name', 'active')
    list_editable = ('active',)

    # do not allow to delete entries in admin
    def has_delete_permission(self, request, obj=None):
        return False

    def get_readonly_fields(self, request, obj=None):
        return obj and ('year',) or ()

    def get_actions(self, request):
        actions = super().get_actions(request)
        if 'delete_selected' in actions:
            del(actions['delete_selected'])
        return actions


class SchoolYearPeriodInlineAdmin(admin.TabularInline):
    model = SchoolYearPeriod
    extra = 0
    min_num = 1


@admin.register(SchoolYearDivision)
class SchoolYearDivisionAdmin(admin.ModelAdmin):
    actions = ('copy_to_school_year',)
    inlines = (SchoolYearPeriodInlineAdmin,)
    list_display = ('name',)
    list_filter = (('school_year', SchoolYearListFilter),)

    def copy_to_school_year(self, request, queryset):
        class SchoolYearForm(forms.Form):
            school_year = forms.ModelChoiceField(
                label=_('Target school year'),
                help_text=_('All selected school year divisions will be copied to selected school year.'),
                queryset=SchoolYear.objects.all(),
            )
        if request.POST.get('post', 'no') == 'yes':
            form = SchoolYearForm(request.POST)
            if form.is_valid():
                school_year = form.cleaned_data['school_year']
                for school_year_division in queryset.all():
                    try:
                        school_year_division.copy_to_school_year(school_year)
                    except IntegrityError:
                        # division already exists for target year
                        pass
                self.message_user(
                    request,
                    _('Selected school year divisions were copied to school year {}.').format(school_year),
                )
                return
        else:
            form = SchoolYearForm()

        adminform = admin.helpers.AdminForm(
            form,
            [(None, {'fields': list(form.base_fields)})],
            {},
            None,
            model_admin=self,
        )

        return render(request, 'leprikon/admin/action_form.html', dict(
            title=_('Select target school year'),
            opts=self.model._meta,
            adminform=adminform,
            media=self.media + adminform.media,
            action='copy_to_school_year',
            action_checkbox_name=admin.helpers.ACTION_CHECKBOX_NAME,
            select_across=request.POST['select_across'],
            selected=request.POST.getlist(admin.helpers.ACTION_CHECKBOX_NAME),
        ))
    copy_to_school_year.short_description = _('Copy selected school year divisions to another school year')

    def get_readonly_fields(self, request, obj=None):
        return ('school_year',) if obj else ()



from django import forms
from django.utils.translation import ugettext_lazy as _

from ..form import FormMixin


class DebtorsForm(FormMixin, forms.Form):
    date = forms.DateField(label=_('Status for date'))



from django.db import models
from django.utils.translation import ugettext_lazy as _


class Citizenship(models.Model):
    name = models.CharField(_('name'), max_length=50)
    order = models.IntegerField(_('order'), blank=True, default=0)
    require_birth_num = models.BooleanField(_('require birth number'), default=True)

    class Meta:
        app_label = 'leprikon'
        ordering = ('order',)
        verbose_name = _('citizenship')
        verbose_name_plural = _('citizenships')

    def __str__(self):
        return self.name



from django import template

from ..rocketchat import get_rc_id, get_rc_url, get_rc_ws_url

register = template.Library()


@register.inclusion_tag('rocketchat/chat.html', takes_context=True)
def chat(context):
    user = getattr(context.get('request'), 'user')
    return {
        'rocketchat_url': get_rc_url(),
        'websocket_url': get_rc_ws_url(),
        'user': user,
        'user_id': user and user.is_authenticated and get_rc_id(user),
    }


@register.inclusion_tag('rocketchat/livechat.html', takes_context=True)
def livechat(context):
    return {
        'rocketchat_url': get_rc_url(),
        'user': getattr(context.get('request'), 'user'),
    }
